#include <cmath>
#include <iostream>
#include <fstream>
#include <cstring>
using namespace std;

const int MAXF = 100;
const int MAXS = 100;
const int MAXP = 100;

enum Tipo {urbana, extraurbana}; 

struct Coord {
  double x;
  double y;
};

struct Fermata {
  int id;
  char indirizzo[MAXS];
  Coord coord;
  Tipo tipo;
};

struct Linea {
  char nome[MAXS];
  int idferm[MAXF];
  int numIdferm; 
};

struct Piano {
  Linea lin[MAXP];
  Fermata fer[MAXF];
  int numLinee;
  int numFermate;
};

/*
 * Funzione ausiliaria che calcola la distanza tra i due punti aventi le coordinate 
 * indicate dai parametri a e b.
 */
double distanza(const Coord a, const Coord b) {
  return sqrt(pow(a.x-b.x, 2) + pow(a.y-b.y, 2));
}


bool caricaPiano(Piano* pp, const char filefermate[], const char filelinee[]) {
  // Due stream: uno per il file delle fermate e uno per quello delle linee.
  // Se ne poteva anche usare uno solo: prima associandolo ad uno dei due file 
  // e chiudendolo al termine delle operazioni di lettura da tale file, poi 
  // associandolo all'altro file
  fstream f1;
  fstream f2;  
  f1.open(filefermate, ios::in);
  f2.open(filelinee, ios::in);

  // Se una delle due operazioni di apertura va male restituisco false e non proseguo
  if(!f1 || !f2)
    return false;

  // Oggetto temporaneo in cui salvo per comodita' i dati di una fermata
  Fermata fr;
  char c;
  // Leggo fino a che non arrivo in fondo al file
  while(f1 >> fr.id >> fr.indirizzo >> fr.coord.x >> fr.coord.y >> c) {
    // a seconda del carattere letto imposto il tipo della fermata
    switch(c) {
      case 'U':
        fr.tipo = urbana;
        break;
      case 'E':
        fr.tipo = extraurbana;
        break;
    }
    // Inserisco la fermata nel piano trasporti
    pp->fer[pp->numFermate++] = fr;
  }
  
  // Oggetto temporaneo in cui salvo per comodita' i dati di una linea 
  Linea l;
  //Leggo fino a che non arrivo in fondo al file
  while(f2 >> l.nome >> l.numIdferm) {
    // Leggo gli id delle fermate
    for(int i=0; i<l.numIdferm; i++) 
      f2 >> l.idferm[i];
    // Inserisco la linea nel piano trasporti
    pp->lin[pp->numLinee++] = l;
  }
  // Chiudo gli stream
  f1.close();
  f2.close();

  return true;
}

/*
 * Funzione ausiliaria che resituisce la fermata avente
 * l'id specificato. Per semplicita' suppongo che una 
 * fermata con tale identificatore sia sicuramente presente
 * nel piano trasporti.
 */ 
Fermata trovaFermata(const Piano* pp, int id) {
  for(int i=0; i<pp->numFermate; i++)
    if(pp->fer[i].id == id)
      return pp->fer[i];
}

/*
 * Calcola la lunghezza del percorso relativo alla linea l
 */
double lunghezzaPercorso(const Piano* pp, Linea l){
  // Lunghezza totale del percorso
  double luntot = 0;
  // Oggetto in cui memorizzo la lunghezza del singolo tratto
  double d;

  for(int i=0; i<l.numIdferm-1; i++) {
      // Prendo la fermata i-esima e quella (i+1)-esima
      Fermata f1 = trovaFermata(pp, l.idferm[i]);
      Fermata f2 = trovaFermata(pp, l.idferm[i+1]);
      // Calcolo la distanza tra le due
      d = distanza(f1.coord, f2.coord);
      // e la sommo alla lunghezza totale
      luntot += d;
  }
  return luntot;
}

bool percorsoPiuLungo(const Piano* pp, char ris[]) {
  // Le distanze sono sicuramente positive quindi posso
  // inizializzare lungmax a zero
  double lungmax = 0;
  // Esamino tutte le linee
  for(int i=0; i<pp->numLinee; i++) {
    // Uso una funzione ausiliaria per calcolare la lunghezza
    // del percorso della i-esima linea
    double d= lunghezzaPercorso(pp, pp->lin[i]);
    // Se e' maggiore di lungmax aggiorno lungmax e scrivo in
    // ris il nome della linea
    if(d > lungmax) {
      strcpy(ris, pp->lin[i].nome);
      lungmax = d;
    }
  }
  // Restituisco true se lungmax e' diverso da zero (c'e' almeno una
  // linea), false altrimenti (il piano trasporti non contiene linee).
  return lungmax != 0;   
}

/*
 * Funzione ausiliaria che mi restituisce l'id della fermata avente 
 * indirizzo ind. Restituisco 0 se non ci sono fermate con tale indirizzo 
 * (gli id sono positivi)
 */
int trovaFermata(const Piano* pp, const char ind[]) {
  for(int i=0; i<pp->numFermate; i++)
    if(strcmp(pp->fer[i].indirizzo, ind) == 0)
      return pp->fer[i].id;
  return 0;
}

void perIndirizzo(const Piano* pp, const char ind[]) {
  // Trovo l'id della fermata che ha indirizzo ind
  int id = trovaFermata(pp, ind);
  // Scorro le linee. Per ogni linea guardo gli id di tutte le fermate e se trovo 
  // quella che mi interessa stampo a video il nome della linea.
  for(int i=0; i<pp->numLinee; i++)
    for(int j=0; j<pp->lin[i].numIdferm; j++)
      if(pp->lin[i].idferm[j] == id)
        cout << pp->lin[i].nome << endl;   
}

/*
 * Funzione ausiliaria che mi dice se la linea l e' interamente contenuta nel 
 * rettangolo individuato da as, bd.
 */
bool contenuta(const Piano* pp, Linea l, const Coord as, const Coord bd) {
  // Scorro tutti gli id delle fermate della linea l
  for(int i=0; i<l.numIdferm; i++) {
      // Recupero l'oggetto fermata corrispondente in quanto mi servono le coordinate
      Fermata f1 = trovaFermata(pp, l.idferm[i]);
      // Se una delle coordinate x o y e' al di fuori del rettangolo restituisco false
      if(f1.coord.x < as.x || f1.coord.x > bd.x || f1.coord.y < bd.y || f1.coord.y > as.y)
        return false;
  }
  // Se non ho trovato fermate esterne al rettangolo, allora vuol 
  // dire che la linea e' interamente contenuta
  return true;
}


int lineeInArea(const Piano* pp, const Coord as, const Coord bd) {
  int conta = 0;
  // Esamino tutte le linee
  for(int i=0; i<pp->numLinee; i++) {
    // Uso una funzione ausiliaria che mi dice se la i-esima linea
    // e' interamente contenuta oppure no
    if(contenuta(pp, pp->lin[i], as, bd))
      conta++;
  }
  return conta;
}

/*
 * Funzione ausiliaria che mi dice se la linea l ferma alla fermata con
 * identificatore id.
 */
bool ferma(const Linea l, int id) {
  for(int i=0; i<l.numIdferm; i++)
    if(l.idferm[i] == id)
      return true;
  return false;
}

void stampa(const Piano* pp) {
  // Esamino tutte le fermate del piano trasporti
  for(int i=0; i<pp->numFermate; i++) {
    // Stampo a video l'indirizzo della i-esima fermata
    cout << pp->fer[i].indirizzo << '\t';
    // Stampo a video il tipo della fermata convertendo l'enumeratore
    // in un carattere 
    cout << (pp->fer[i].tipo == urbana ? 'U' : 'E') << '\t';
    // Per trovare le linee che fermano in tale fermata le esamino tutte
    // e adopero una funzione ausiliaria per verificare se la j-esima linea
    // ferma nella fermata che mi interessa
    for(int j=0; j<pp->numLinee; j++)
      if(ferma(pp->lin[j], pp->fer[i].id))
        cout << pp->lin[j].nome << '\t';
    cout << endl;
  }
}

